@@ -17,6 +17,7 @@
 #include <QObject>
 #include <QStandardPaths>
 #include <QTest>
+#include <QPalette>
 
 #include <algorithm>
 
@@ -234,6 +235,7 @@ private Q_SLOTS:
                                                      QStringLiteral("Mustache/Handlebars (HTML)/PHP"),
                                                      QStringLiteral("Doxygen"),
                                                      QStringLiteral("Modelines"),
+                                                     QStringLiteral("SPDX-Comments"),
                                                      QStringLiteral("HTML"),
                                                      QStringLiteral("CSS"),
                                                      QStringLiteral("SQL (MySQL)"),
@@ -276,9 +278,12 @@ private Q_SLOTS:
         }
 
         // ensure all ids are there from 1..size
-        for (int i = 1; i <= formatIds.size(); ++i) {
-            QVERIFY(formatIds.contains(i));
-        }
+        // this is no longer feasible, as we e.g. skip the "Comments"
+        // syntax definition in includedDefinitions as it is not contributing
+        // any context/rule
+        //for (int i = 1; i <= formatIds.size(); ++i) {
+        //    QVERIFY(formatIds.contains(i));
+        //}
     }
 
     void testReload()
@@ -531,6 +536,16 @@ private Q_SLOTS:
         QVERIFY(lists.contains(unexistedName) == false);
         QVERIFY(def.setKeywordList(unexistedName, QStringList()) == false);
     }
+
+    void testAutomaticThemeSelection()
+    {
+        QPalette palette;
+        palette.setColor(QPalette::Base, QColor(27, 30, 32));
+        palette.setColor(QPalette::Highlight, QColor(61, 174, 253));
+        auto theme = m_repo.themeForPalette(palette);
+        QCOMPARE(theme.name(), QStringLiteral("Breeze Dark"));
+    }
+
 };
 }
 
